'''#!/usr/bin/env python3'''
#!/home/robobib/robobin_ws/venv/bin/python

import rclpy
from rclpy.node import Node
from sensor_msgs.msg import Imu
import smbus
import math
from ahrs.filters import Madgwick


class IMUNode(Node):
    def __init__(self):
        super().__init__("imusd_node")
        self.get_logger().info("hello")


        self.publisher = self.create_publisher(Imu, 'imu/data_raw',10)
    
        self.bus = smbus.SMBus(1)
        self.DEVICE_ADDRESS = 0x68  # MPU6050 device address
        self.mpu_init()

        self.filter = Madgwick()
        self.orientation = [0.0, 0.0, 0.0, 1.0]



        time_period = 0.01
        self.timer = self.create_timer(time_period, self.timer_callback)



    def mpu_init(self):
        PWR_MGMT_1 = 0x6B
        SMPLRT_DIV = 0x19
        CONFIG = 0x1A
        GYRO_CONFIG = 0x1B
        ACCEL_CONFIG = 0x1C
        INT_ENABLE = 0x38

        # Write to sample rate register
        self.bus.write_byte_data(self.DEVICE_ADDRESS, SMPLRT_DIV, 7)
        # Write to power management register to wake up the MPU6050
        self.bus.write_byte_data(self.DEVICE_ADDRESS, PWR_MGMT_1, 1)
        # Write to configuration register
        self.bus.write_byte_data(self.DEVICE_ADDRESS, CONFIG, 0)
        # Write to gyroscope configuration register
        self.bus.write_byte_data(self.DEVICE_ADDRESS, GYRO_CONFIG, 24)
        # Write to accelerometer configuration register
        self.bus.write_byte_data(self.DEVICE_ADDRESS, ACCEL_CONFIG, 0)
        # Enable interrupts
        self.bus.write_byte_data(self.DEVICE_ADDRESS, INT_ENABLE, 1)

    def read_raw_data(self, addr):
        # Read two bytes of data from the given address
        high = self.bus.read_byte_data(self.DEVICE_ADDRESS, addr)
        low = self.bus.read_byte_data(self.DEVICE_ADDRESS, addr+1)
        # Combine higher and lower bytes
        value = (high << 8) | low
        # Convert to signed integer
        if value > 32768:
            value = value - 65536

        return value

    def timer_callback(self):


        ACCEL_XOUT_H = 0x3B
        ACCEL_YOUT_H = 0x3D
        ACCEL_ZOUT_H = 0x3F

        GYRO_XOUT_H = 0x43
        GYRO_YOUT_H = 0x45
        GYRO_ZOUT_H = 0x47

        # Read accelerometer data
        acc_x = self.read_raw_data(ACCEL_XOUT_H)
        acc_y = self.read_raw_data(ACCEL_YOUT_H)
        acc_z = self.read_raw_data(ACCEL_ZOUT_H)
        
        # Read gyroscope data
        gyro_x = self.read_raw_data(GYRO_XOUT_H)
        gyro_y = self.read_raw_data(GYRO_YOUT_H)
        gyro_z = self.read_raw_data(GYRO_ZOUT_H)


        ax_offset = 0.01467432
        ay_offset = -0.00603101
        az_offset = 0.06171924

        gx_offset = -0.17447328
        gy_offset = 0.08720611
        gz_offset = 0.16423664

        Ax = (acc_x / 16384.0)  - (ax_offset) # Accelerometer sensitivity scale factor
        Ay = (acc_y / 16384.0)  - (ay_offset) 
        Az = (acc_z / 16384.0) - (az_offset) 
        Ax = 9.81*Ax
        Ay = 9.81*Ay
        Az = 9.81*Az
        
        Gx = (gyro_x / 131.0) - (gx_offset)   # Gyroscope sensitivity scale factor
        Gy = (gyro_y / 131.0) - (gy_offset) 
        Gz = (gyro_z / 131.0)  - (gz_offset) 
        
        Gx = Gx/180 *math.pi
        Gy = Gy/180 *math.pi
        Gz = Gz/180 *math.pi

        self.orientation = self.filter.updateIMU(self.orientation,
                                                 [Gx, Gy, Gz],
                                                 [Ax/9.81, Ay/9.81, Az/9.81])
        


        imu_msg = Imu()
        imu_msg.header.stamp =self.get_clock().now().to_msg()
        imu_msg.header.frame_id = 'imu_link'
        imu_msg.orientation_covariance[0] = -1.0

        imu_msg.angular_velocity.x = Gx
        imu_msg.angular_velocity.y = Gy
        imu_msg.angular_velocity.z = Gz

        imu_msg.linear_acceleration.x = Ax
        imu_msg.linear_acceleration.y = Ay
        imu_msg.linear_acceleration.z = Az

        imu_msg.orientation.x =self.orientation[0]
        imu_msg.orientation.y =self.orientation[1]
        imu_msg.orientation.z =self.orientation[2]
        imu_msg.orientation.w =self.orientation[3]

        imu_msg.orientation_covariance = [0.0025, 0, 0,
                                          0, 0.0025, 0,
                                          0, 0, 0.0025]

        self.publisher.publish(imu_msg)

        





def main(args=None):
    rclpy.init(args=args)
    imu_node = IMUNode()
    rclpy.spin(imu_node)
    rclpy.shutdown()

if __name__ == '__main__':
    main()