#!/usr/bin/env python3
'''#!/home/robobib/robobin_ws/venv/bin/python'''

import rclpy
from rclpy.node import Node
from sensor_msgs.msg import Imu
import smbus
import math
import numpy as np



class IMUNode(Node):
    def __init__(self):
        super().__init__("imusd_node")
        self.get_logger().info("hello")


        self.publisher = self.create_publisher(Imu, 'imu/data_raw',10)
    
        self.bus = smbus.SMBus(1)
        self.DEVICE_ADDRESS = 0x68  # MPU6050 device address
        self.mpu_init()

        self.dt = 0.01  # Time step (same as timer period)
        self.roll = 0.0
        self.pitch = 0.0
        self.yaw = 0.0
        self.alpha = 0.98  # Filter coefficient


        time_period = 0.01
        self.timer = self.create_timer(time_period, self.timer_callback)



    def mpu_init(self):
        PWR_MGMT_1 = 0x6B
        SMPLRT_DIV = 0x19
        CONFIG = 0x1A
        GYRO_CONFIG = 0x1B
        ACCEL_CONFIG = 0x1C
        INT_ENABLE = 0x38

        # Write to sample rate register
        self.bus.write_byte_data(self.DEVICE_ADDRESS, SMPLRT_DIV, 7)
        # Write to power management register to wake up the MPU6050
        self.bus.write_byte_data(self.DEVICE_ADDRESS, PWR_MGMT_1, 1)
        # Write to configuration register
        self.bus.write_byte_data(self.DEVICE_ADDRESS, CONFIG, 0)
        # Write to gyroscope configuration register
        self.bus.write_byte_data(self.DEVICE_ADDRESS, GYRO_CONFIG, 24)
        # Write to accelerometer configuration register
        self.bus.write_byte_data(self.DEVICE_ADDRESS, ACCEL_CONFIG, 0)
        # Enable interrupts
        self.bus.write_byte_data(self.DEVICE_ADDRESS, INT_ENABLE, 1)

    def read_raw_data(self, addr):
        # Read two bytes of data from the given address
        high = self.bus.read_byte_data(self.DEVICE_ADDRESS, addr)
        low = self.bus.read_byte_data(self.DEVICE_ADDRESS, addr+1)
        # Combine higher and lower bytes
        value = (high << 8) | low
        # Convert to signed integer
        if value > 32768:
            value = value - 65536

        return value
    
    def euler_to_quaternion(self, roll, pitch, yaw):
        """
        Convert an Euler angle to a quaternion.

        Input
          :param roll: Rotation around the X-axis in radians
          :param pitch: Rotation around the Y-axis in radians
          :param yaw: Rotation around the Z-axis in radians

        Output
          :return qx, qy, qz, qw: Quaternion components
        """
        qx = math.sin(roll/2) * math.cos(pitch/2) * math.cos(yaw/2) - \
             math.cos(roll/2) * math.sin(pitch/2) * math.sin(yaw/2)
        qy = math.cos(roll/2) * math.sin(pitch/2) * math.cos(yaw/2) + \
             math.sin(roll/2) * math.cos(pitch/2) * math.sin(yaw/2)
        qz = math.cos(roll/2) * math.cos(pitch/2) * math.sin(yaw/2) - \
             math.sin(roll/2) * math.sin(pitch/2) * math.cos(yaw/2)
        qw = math.cos(roll/2) * math.cos(pitch/2) * math.cos(yaw/2) + \
             math.sin(roll/2) * math.sin(pitch/2) * math.sin(yaw/2)

        return qx, qy, qz, qw

    def complementary_filter(self, acc, gyr, dt):
        # Integrate gyroscope data
        self.roll += gyr[0] * dt
        self.pitch += gyr[1] * dt
        self.yaw += gyr[2] * dt  # Yaw accumulates gyro data

        # Calculate accelerometer angles
        roll_acc = math.atan2(acc[1], acc[2])
        pitch_acc = math.atan2(-acc[0], math.sqrt(acc[1]**2 + acc[2]**2))

        # Apply complementary filter
        self.roll = self.alpha * self.roll + (1 - self.alpha) * roll_acc
        self.pitch = self.alpha * self.pitch + (1 - self.alpha) * pitch_acc
        # Yaw remains as gyroscope integration

        return self.roll, self.pitch, self.yaw

    def timer_callback(self):


        ACCEL_XOUT_H = 0x3B
        ACCEL_YOUT_H = 0x3D
        ACCEL_ZOUT_H = 0x3F

        GYRO_XOUT_H = 0x43
        GYRO_YOUT_H = 0x45
        GYRO_ZOUT_H = 0x47

        # Read accelerometer data
        acc_x = self.read_raw_data(ACCEL_XOUT_H)
        acc_y = self.read_raw_data(ACCEL_YOUT_H)
        acc_z = self.read_raw_data(ACCEL_ZOUT_H)
        
        # Read gyroscope data
        gyro_x = self.read_raw_data(GYRO_XOUT_H)
        gyro_y = self.read_raw_data(GYRO_YOUT_H)
        gyro_z = self.read_raw_data(GYRO_ZOUT_H)


        ax_offset = 0.01467432
        ay_offset = -0.00603101
        az_offset = 0.06171924

        gx_offset = -0.17447328
        gy_offset = 0.08720611
        gz_offset = 0.16423664

        Ax = (acc_x / 16384.0)  - (ax_offset) # Accelerometer sensitivity scale factor
        Ay = (acc_y / 16384.0)  - (ay_offset) 
        Az = (acc_z / 16384.0) - (az_offset) 
        Ax = 9.81*Ax
        Ay = 9.81*Ay
        Az = 9.81*Az
        
        Gx = (gyro_x / 131.0) - (gx_offset)   # Gyroscope sensitivity scale factor
        Gy = (gyro_y / 131.0) - (gy_offset) 
        Gz = (gyro_z / 131.0)  - (gz_offset) 
        
        Gx = Gx/180 *math.pi
        Gy = Gy/180 *math.pi
        Gz = Gz/180 *math.pi

        # Prepare data arrays
        acc = np.array([Ax, Ay, Az])
        gyr = np.array([Gx, Gy, Gz])

        # Estimate orientation using the complementary filter
        roll, pitch, yaw = self.complementary_filter(acc, gyr, self.dt)

        # Convert roll, pitch, yaw to quaternion
        qx, qy, qz, qw = self.euler_to_quaternion(roll, pitch, yaw)
        


        imu_msg = Imu()
        imu_msg.header.stamp =self.get_clock().now().to_msg()
        imu_msg.header.frame_id = 'imu_link'
        imu_msg.orientation_covariance[0] = -1.0

        imu_msg.angular_velocity.x = Gx
        imu_msg.angular_velocity.y = Gy
        imu_msg.angular_velocity.z = Gz

        imu_msg.linear_acceleration.x = Ax
        imu_msg.linear_acceleration.y = Ay
        imu_msg.linear_acceleration.z = Az

        imu_msg.orientation.x = qx
        imu_msg.orientation.y = qy
        imu_msg.orientation.z = qz
        imu_msg.orientation.w = qw
        imu_msg.orientation_covariance = [0.01, 0, 0,
                                          0, 0.01, 0,
                                          0, 0, 0.05] 

        self.publisher.publish(imu_msg)

        





def main(args=None):
    rclpy.init(args=args)
    imu_node = IMUNode()
    rclpy.spin(imu_node)
    rclpy.shutdown()

if __name__ == '__main__':
    main()